from pydantic import BaseModel
import asyncio
from reaktion_next.events import OutEvent


class AtomTransport(BaseModel):
    queue: asyncio.Queue

    async def put(self, event: OutEvent):
        await self.queue.put(event)

    async def get(self) -> OutEvent:
        return await self.queue.get()

    class Config:
        arbitrary_types_allowed = True


class MockTransport(AtomTransport):
    async def get(self, timeout=3) -> OutEvent:
        return await asyncio.wait_for(self.queue.get(), timeout=timeout)

    pass



import asyncio
from typing import Awaitable, Callable, Optional
from pydantic import BaseModel, Field
from fluss_next.api.schema import BaseGraphNodeFragmentBase
from reaktion_next.atoms.errors import AtomQueueFull
from reaktion_next.events import EventType, InEvent, OutEvent
import logging
from rekuest_next.actors.types import Assignment
from reaktion_next.atoms.transport import AtomTransport
from typing import Dict, Any

logger = logging.getLogger(__name__)


class Atom(BaseModel):
    node: BaseGraphNodeFragmentBase
    transport: AtomTransport
    alog: Optional[Callable[[str, str, str], Awaitable[None]]] = Field(exclude=True)
    globals: Dict[str, Any] = Field(default_factory=dict)
    assignment: Assignment

    _private_queue: asyncio.Queue = None

    async def run(self):
        raise NotImplementedError("This needs to be implemented")

    async def get(self) -> InEvent:
        assert self._private_queue is not None, "Atom not started"
        return await self._private_queue.get()

    async def put(self, event: InEvent):
        assert self._private_queue is not None, "Atom not started"
        try:
            logger.info(f"Putting event {event}")
            await self._private_queue.put(event)  # TODO: Make put no wait?
        except asyncio.QueueFull as e:
            logger.error(f"{self.node.id} private queue is full")
            raise AtomQueueFull(f"{self.node.id} private queue is full") from e
        except Exception as e:
            logger.error(f"{self.node.id} FAILED", exc_info=True)
            print(f"{self.node.id} FAILED", e)
            await self.transport.put(
                OutEvent(
                    handle="return_0",
                    type=EventType.ERROR,
                    source=self.node.id,
                    value=e,
                    caused_by=[-1],
                )
            )

    async def aenter(self):
        self._private_queue = asyncio.Queue()

    async def aexit(self):
        self._private_queue = None

    async def __aenter__(self):
        await self.aenter()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aexit()

    async def start(self):
        try:
            await self.run()
        except Exception as e:
            logger.error(f"{self.node.id} FAILED", exc_info=True)
            await self.transport.put(
                OutEvent(
                    handle="return_0",
                    type=EventType.ERROR,
                    source=self.node.id,
                    value=e,
                    caused_by=[-1],
                )
            )

    @property
    def set_values(self) -> Dict[str, Any]:
        defaults = self.node.constants_map or {}
        my_globals = self.globals or {}
        return {**defaults, **my_globals}

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



import asyncio
from reaktion_next.events import OutEvent, Returns, EventType, InEvent
from reaktion_next.atoms.base import Atom
import logging
from pydantic import Field
from typing import Dict, List
import asyncio

logger = logging.getLogger(__name__)


class FilterAtom(Atom):
    async def filter(self, event: InEvent) -> bool:
        raise NotImplementedError("This needs to be implemented")

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    try:
                        result = await self.filter(event)
                        if result is True:
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=event.value,
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )
                    except Exception as e:
                        logger.error(f"{self.node.id} map failed", exc_info=True)
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                source=self.node.id,
                                value=e,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            logger.debug(f"Atom {self.node} is getting cancelled")
            raise e


class MapAtom(Atom):
    async def map(self, event: InEvent) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    try:
                        result = await self.map(event)
                        if result is None:
                            value = ()
                        elif isinstance(result, list) or isinstance(result, tuple):
                            value = result
                        else:
                            value = (result,)

                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                value=value,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    except Exception as e:
                        logger.error(f"{self.node.id} map failed", exc_info=True)
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                source=self.node.id,
                                value=e,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            logger.debug(f"Atom {self.node} is getting cancelled")
            raise e


class MergeMapAtom(Atom):
    async def merge_map(self, event: InEvent) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def run(self):
        try:
            while True:
                event = await self.get()
                print(event)

                if event.type == EventType.NEXT:
                    try:
                        async for result in self.merge_map(event):
                            if result is None:
                                value = ()
                            elif isinstance(result, list) or isinstance(result, tuple):
                                value = result
                            else:
                                value = (result,)
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=value,
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )

                    except Exception as e:
                        logger.error(f"{self.node.id} map failed")
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                source=self.node.id,
                                value=e,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            logger.debug(f"Atom {self.node} is getting cancelled")
            raise e


class OrderedAtom(Atom):
    runningEvents: Dict[int, asyncio.Task] = Field(default_factory=dict)
    publish_queue: List[int] = Field(default_factory=list)

    async def map(self, event: InEvent) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def check_ordered(self):
        tasks_to_remove = []
        for key, task in self.runningEvents.items():
            if task.done():
                exception = task.exception()
                if exception:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=exception,
                            source=self.node.id,
                            caused_by=[key],
                        )
                    )
                else:
                    if key in self.publish_queue:
                        if self.publish_queue[0] == key:
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=task.result(),
                                    source=self.node.id,
                                    caused_by=[key],
                                )
                            )
                            self.publish_queue.pop(0)
                            tasks_to_remove.append(key)

        for key in tasks_to_remove:
            self.runningEvents.pop(key)

    async def publish_changes(self):
        while True:
            await asyncio.sleep(0.1)
            await self.check_ordered()

    async def run(self):
        publish_task = asyncio.create_task(self.publish_changes())

        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    self.publish_queue.append(event.current_t)
                    self.runningEvents[event.current_t] = asyncio.create_task(
                        self.map(event)
                    )

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await asyncio.gather(*self.runningEvents.values())
                    await self.check_ordered()

                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    for key, value in self.runningEvents.items():
                        value.cancel()
                        try:
                            await value
                        except asyncio.CancelledError:
                            pass

                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass

                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            publish_task.cancel()
            try:
                await publish_task
            except asyncio.CancelledError:
                pass

            logger.debug(f"Atom {self.node} is getting cancelled")
            raise e


class AsCompletedAtom(Atom):
    runningEvents: Dict[int, asyncio.Task] = Field(default_factory=dict)

    async def map(self, event: InEvent) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def check_as_completed(self):
        tasks_to_remove = []
        for key, task in self.runningEvents.items():
            if task.done():
                exception = task.exception()
                if exception:
                    logger.error(f"{self.node.id} map failed with {exception}")
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=exception,
                            source=self.node.id,
                            caused_by=[key],
                        )
                    )

                else:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.NEXT,
                            value=task.result(),
                            source=self.node.id,
                            caused_by=[key],
                        )
                    )
                tasks_to_remove.append(key)

        for key in tasks_to_remove:
            self.runningEvents.pop(key)

    async def publish_changes(self):
        while True:
            await asyncio.sleep(0.1)
            await self.check_as_completed()

    async def run(self):
        publish_task = asyncio.create_task(self.publish_changes())

        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    self.runningEvents[event.current_t] = asyncio.create_task(
                        self.map(event)
                    )

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await asyncio.gather(*self.runningEvents.values())
                    await self.check_as_completed()
                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    for key, value in self.runningEvents.items():
                        value.cancel()
                        try:
                            await value
                        except asyncio.CancelledError:
                            pass

                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass

                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            publish_task.cancel()
            try:
                await publish_task
            except asyncio.CancelledError:
                pass

            logger.debug(f"Atom {self.node} is getting cancelled")
            raise e



import logging
from typing import Callable, Dict, Optional
import asyncio
from pydantic import BaseModel, Field

from fluss_next.api.schema import (
    ArgNodeFragment,
    RekuestNodeFragmentBase,
    FlowFragment,
    ReactiveNodeFragment,
    RekuestFilterNodeFragment,
    ReturnNodeFragment,
    acreate_run,
    asnapshot,
    atrack,
)
from reaktion_next.atoms.transport import AtomTransport

from reaktion_next.atoms.utils import atomify
from reaktion_next.contractors import NodeContractor, arkicontractor
from reaktion_next.events import EventType, InEvent, OutEvent

from reaktion_next.utils import connected_events
from rekuest_next.actors.base import Actor
from rekuest_next.api.schema import (
    AssignationEventKind,
    ReservationFragment,
    NodeFragment,
)
from rekuest_next.postmans.utils import RPCContract, ContractStatus
from typing import Any
from rekuest_next.collection.collector import AssignationCollector
from rekuest_next.actors.transport.types import AssignTransport
from rekuest_next.actors.types import Assignment, Passport

logger = logging.getLogger(__name__)


class NodeState(BaseModel):
    latestevent: OutEvent


class FlowActor(Actor):
    definition: NodeFragment
    is_generator: bool = False
    flow: FlowFragment
    agent: Any
    contracts: Dict[str, RPCContract] = Field(default_factory=dict)
    expand_inputs: bool = False
    shrink_outputs: bool = False
    provided = False
    arkitekt_contractor: NodeContractor = arkicontractor
    snapshot_interval: int = 40
    condition_snapshot_interval: int = 40
    contract_states: Dict[str, ContractStatus] = Field(default_factory=dict)
    contract_t: int = 0

    # Functionality for running the flow

    # Assign Related Functionality
    run_mutation: Callable = acreate_run
    snapshot_mutation: Callable = asnapshot
    track_mutation: Callable = atrack

    atomifier: Callable = atomify
    """ Atomifier is a function that takes a node and returns an atom """

    run_states: Dict[
        str,
        Dict[str, NodeState],
    ] = Field(default_factory=dict)

    reservation_state: Dict[str, ReservationFragment] = Field(default_factory=dict)
    _lock: Optional[asyncio.Lock] = None
    _condition = None

    async def on_provide(self, passport: Passport):
        self._lock = asyncio.Lock()

    async def on_local_log(self, reference, *args, **kwargs):
        logger.log(f"Contract log for {reference} {args} {kwargs}")

    async def on_assign(
        self,
        assignment: Assignment,
        collector: AssignationCollector,
        transport: AssignTransport,
    ):
        try:
            rekuest_nodes = [
                x
                for x in self.flow.graph.nodes
                if isinstance(x, RekuestNodeFragmentBase)
            ]

            rekuest_contracts = {
                node.id: await self.arkitekt_contractor(node, self)
                for node in rekuest_nodes
            }

            self.contracts = {**rekuest_contracts}
            futures = [contract.aenter() for contract in self.contracts.values()]
            await asyncio.gather(*futures)

            run = await self.run_mutation(
                assignation=assignment.assignation,
                flow=self.flow,
                snapshot_interval=self.snapshot_interval,
            )
            print(self.is_generator)

            t = 0
            state = {}
            await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

            event_queue = asyncio.Queue()

            atomtransport = AtomTransport(queue=event_queue)

            argNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)
            ][0]
            returnNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)
            ][0]

            participatingNodes = [
                x
                for x in self.flow.graph.nodes
                if isinstance(x, RekuestNodeFragmentBase)
                or isinstance(x, ReactiveNodeFragment)
            ]

            stream = argNode.outs[0]
            stream_keys = []
            for i in stream:
                stream_keys.append(i.key)

            return_stream = returnNode.outs[0]

            global_keys = []
            for i in self.flow.graph.globals:
                global_keys.append(i.port.key)

            globalMap: Dict[str, Dict[str, Any]] = {}
            streamMap: Dict[str, Any] = {}

            for port in self.definition.args:
                if port.key in stream_keys:
                    streamMap[port.key] = assignment.args[port.key]
                if port.key in global_keys:
                    for i in self.flow.graph.globals:
                        if i.port.key == port.key:
                            for map in i.to_keys:
                                nodeid, key = map.split(".")
                                globalMap.setdefault(nodeid, {})[key] = assignment.args[
                                    port.key
                                ]

            atoms = {
                x.id: self.atomifier(
                    x,
                    atomtransport,
                    self.contracts.get(x.id, None),
                    globalMap.get(x.id, {}),
                    assignment,
                    alog=transport.log_event,
                )
                for x in participatingNodes
            }

            await asyncio.gather(*[atom.aenter() for atom in atoms.values()])
            tasks = [asyncio.create_task(atom.start()) for atom in atoms.values()]
            logger.info("Starting all Atoms")
            value = [streamMap[key] for key in stream_keys]

            initial_event = OutEvent(
                handle="return_0",
                type=EventType.NEXT,
                source=argNode.id,
                value=value,
                caused_by=[t],
            )
            initial_done_event = OutEvent(
                handle="return_0",
                type=EventType.COMPLETE,
                source=argNode.id,
                caused_by=[t],
            )

            logger.info(f"Putting initial event {initial_event}")

            await event_queue.put(initial_event)
            await event_queue.put(initial_done_event)

            edge_targets = [e.target for e in self.flow.graph.edges]
            nodes_without_instream = [
                x
                for x in participatingNodes
                if len(x.ins[0]) == 0 and x.id not in edge_targets
            ]

            for node in nodes_without_instream:
                assert node.id in atoms, "Atom not found. Should not happen."
                atom = atoms[node.id]

                initial_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.NEXT,
                    value=[],
                    current_t=t,
                )
                done_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.COMPLETE,
                    current_t=t,
                )

                await atom.put(initial_event)
                await atom.put(done_event)

            complete = False

            returns = []

            while not complete:
                event: OutEvent = await event_queue.get()
                event_queue.task_done()
                print(event)

                if event.type == EventType.ERROR:
                    # raise event.value
                    pass

                track = await self.track_mutation(
                    reference=event.source + "_track_" + str(t),
                    run=run,
                    source=event.source,
                    handle=event.handle,
                    caused_by=event.caused_by,
                    value=event.value
                    if event.value and not isinstance(event.value, Exception)
                    else str(event.value),
                    kind=event.type,
                    t=t,
                )
                state[event.source] = track.id

                # We tracked the events and proceed

                if t % self.snapshot_interval == 0:
                    await self.snapshot_mutation(
                        run=run, events=list(state.values()), t=t
                    )

                # Creat new events with the new timepoint
                spawned_events = connected_events(self.flow.graph, event, t)
                # Increment timepoint
                t += 1
                # needs to be the old one for now
                if not spawned_events:
                    logger.warning(f"No events spawned from {event}")

                for spawned_event in spawned_events:
                    logger.info(f"-> {spawned_event}")

                    if spawned_event.target == returnNode.id:
                        if spawned_event.type == EventType.NEXT:
                            yield_dict = {}

                            for port, value in zip(return_stream, spawned_event.value):
                                yield_dict[port.key] = value

                            await transport.log_event(
                                kind=AssignationEventKind.YIELD,
                                returns=yield_dict,
                            )

                        if spawned_event.type == EventType.ERROR:
                            await self.snapshot_mutation(
                                run=run, events=list(state.values()), t=t
                            )
                            raise spawned_event.value

                        if spawned_event.type == EventType.COMPLETE:
                            await self.snapshot_mutation(
                                run=run, events=list(state.values()), t=t
                            )
                            await transport.log_event(
                                kind=AssignationEventKind.DONE,
                                message="Done ! :)",
                            )

                            logger.info("Done ! :)")

                    else:
                        assert (
                            spawned_event.target in atoms
                        ), "Unknown target. Your flow is connected wrong"
                        if spawned_event.target in atoms:
                            await atoms[spawned_event.target].put(spawned_event)

            for task in tasks:
                task.cancel()

            await asyncio.gather(*tasks, return_exceptions=True)
            logging.info("Collecting...")
            await self.collector.collect(assignment.id)
            logging.info("Done ! :)")

        except asyncio.CancelledError:
            for task in tasks:
                task.cancel()
            await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

            try:
                await asyncio.wait_for(
                    asyncio.gather(*tasks, return_exceptions=True), timeout=4
                )
            except asyncio.TimeoutError:
                pass

            await self.collector.collect(assignment.id)
            await transport.log_event(
                kind=AssignationEventKind.CANCELLED, message="Cancelled"
            )

        except Exception as e:
            print(e)
            logging.critical(f"Assignation {assignment} failed", exc_info=True)
            await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

            await self.collector.collect(assignment.id)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=repr(e),
            )

    async def on_unprovide(self):
        for contract in self.contracts.values():
            await contract.aexit()



import asyncio
from typing import List, Tuple, Optional
from reaktion_next.atoms.combination.base import CombinationAtom
from reaktion_next.events import EventType, OutEvent, InEvent
import logging
from pydantic import Field
from reaktion_next.atoms.helpers import index_for_handle
import functools

logger = logging.getLogger(__name__)


class ZipAtom(CombinationAtom):
    state: List[Optional[InEvent]] = Field(default_factory=lambda: [None, None])
    complete: List[Optional[InEvent]] = Field(default_factory=lambda: [None, None])

    async def run(self):
        self.state = list(map(lambda x: None, self.node.instream))
        self.complete = list(map(lambda x: None, self.node.instream))

        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.COMPLETE:
                    self.complete[index_for_handle(event.handle)] = event
                    if all(map(lambda x: x is not None, self.complete)):
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.COMPLETE,
                                source=self.node.id,
                                caused_by=map(lambda x: x.current_t, self.complete),
                            )
                        )
                        break

                if event.type == EventType.NEXT:
                    self.state[index_for_handle(event.handle)] = event
                    if all(map(lambda x: x is not None, self.state)):
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                source=self.node.id,
                                value=functools.reduce(
                                    lambda a, b: a + b.value, self.state, tuple()
                                ),
                                caused_by=map(lambda x: x.current_t, self.state),
                            )
                        )
                        # Reinitalizing
                        self.state = list(map(lambda x: None, self.node.instream))

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception:
            logger.exception(f"Atom {self.node} excepted")



import asyncio

from typing import Any, List, Optional
from rekuest_next.postmans.utils import RPCContract
from reaktion_next.atoms.helpers import node_to_reference

from fluss_next.api.schema import RekuestMapNodeFragment

from reaktion_next.atoms.generic import (
    MapAtom,
    MergeMapAtom,
    AsCompletedAtom,
    OrderedAtom,
)
from reaktion_next.events import InEvent
import logging
from rekuest_next.utils import ReservationContext


logger = logging.getLogger(__name__)


class ArkitektMapAtom(MapAtom):
    node: RekuestMapNodeFragment
    contract: ReservationContext

    async def map(self, event: InEvent) -> Optional[List[Any]]:
        kwargs = self.set_values

        stream_one = self.node.ins[0]
        for arg, item in zip(event.value, stream_one):
            kwargs[item.key] = arg

        returns = await self.contract.acall_raw(
            parent=self.assignment.assignation,
            reference=node_to_reference(self.node, event),
            **kwargs,
        )

        out = []
        stream_one = self.node.outs[0]
        for arg in stream_one:
            out.append(returns[arg.key])

        return out
        # return await self.contract.aassign(*args)


class ArkitektMergeMapAtom(MergeMapAtom):
    node: RekuestMapNodeFragment
    contract: ReservationContext

    async def merge_map(self, event: InEvent) -> Optional[List[Any]]:
        print("merge_map", event)
        kwargs = self.set_values

        stream_one = self.node.ins[0]
        for arg, item in zip(event.value, stream_one):
            kwargs[item.key] = arg

        print(kwargs, kwargs)

        async for r in self.contract.aiterate_raw(
            parent=self.assignment.assignation,
            reference=node_to_reference(self.node, event),
            **kwargs,
        ):
            print(r)
            out = []
            stream_one = self.node.outs[0]
            for arg in stream_one:
                out.append(r[arg.key])

            yield out


class ArkitektAsCompletedAtom(AsCompletedAtom):
    node: RekuestMapNodeFragment
    contract: ReservationContext

    async def map(self, event: InEvent) -> Optional[List[Any]]:
        kwargs = self.set_values

        stream_one = self.node.instream[0]
        for arg, item in zip(event.value, stream_one):
            kwargs[item.key] = arg

        returns = await self.contract.aassign_retry(
            kwargs=kwargs,
            parent=self.assignment,
            reference=node_to_reference(self.node, event),
        )

        out = []
        stream_one = self.node.outstream[0]
        for arg in stream_one:
            out.append(returns[arg.key])

        return out


class ArkitektOrderedAtom(OrderedAtom):
    node: RekuestMapNodeFragment
    contract: ReservationContext

    async def map(self, event: InEvent) -> Optional[List[Any]]:
        kwargs = self.set_values

        stream_one = self.node.instream[0]
        for arg, item in zip(event.value, stream_one):
            kwargs[item.key] = arg

        returns = await self.contract.aassign_retry(
            kwargs=kwargs,
            parent=self.assignment,
            reference=node_to_reference(self.node, event),
        )

        out = []
        stream_one = self.node.outstream[0]
        for arg in stream_one:
            out.append(returns[arg.key])

        return out



import asyncio
from typing import List, Optional
from reaktion_next.atoms.helpers import index_for_handle
from reaktion_next.atoms.combination.base import CombinationAtom
from reaktion_next.events import EventType, OutEvent, InEvent
import logging
import functools
import asyncio
from typing import Optional
from pydantic import Field

logger = logging.getLogger(__name__)


class GateAtom(CombinationAtom):
    buffer: Optional[asyncio.Queue] = None

    async def run(self):
        self.buffer = asyncio.Queue()

        forward_first = self.set_values.get("forward_first", True)

        try:
            while True:
                event = await self.get()
                logger.info(f"GateAtom {self.node} received {event}")

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                streamIndex = index_for_handle(event.handle)

                if event.type == EventType.COMPLETE:
                    if streamIndex == 0:
                        if forward_first:
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.COMPLETE,
                                    value=event.value,
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )
                            forward_first = False

                        else:
                            await self.buffer.put(event)

                if event.type == EventType.NEXT:
                    if streamIndex == 0:
                        if forward_first:
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=event.value,
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )
                            forward_first = False

                        else:
                            logger.info("Buffering event")
                            await self.buffer.put(event)

                    else:
                        if self.buffer.empty():
                            forward_first = True
                            logger.info("Buffer is empty, waiting for first event")
                        else:
                            get_event = await self.buffer.get()
                            if get_event.type == EventType.COMPLETE:
                                await self.transport.put(
                                    OutEvent(
                                        handle="return_0",
                                        type=EventType.COMPLETE,
                                        source=self.node.id,
                                        caused_by=[
                                            get_event.current_t,
                                            event.current_t,
                                        ],
                                    )
                                )
                                break
                            else:
                                await self.transport.put(
                                    OutEvent(
                                        handle="return_0",
                                        type=EventType.NEXT,
                                        value=get_event.value,
                                        source=self.node.id,
                                        caused_by=[
                                            get_event.current_t,
                                            event.current_t,
                                        ],
                                    )
                                )

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception:
            logger.exception(f"Atom {self.node} excepted")



from reaktion_next.actor import FlowActor
from rekuest_next.agents.base import BaseAgent
from rekuest_next.actors.reactive.api import useInstanceID
import logging
from rekuest_next.register import register_func
from rekuest_next.actors.base import Actor
from rekuest_next.actors.types import Passport, Assignment
from fluss_next.api.schema import RekuestNodeFragmentBase, aget_flow
from rekuest_next.api.schema import NodeKind

from typing import Optional
from rekuest_next.api.schema import (
    PortInput,
    DependencyInput,
    DefinitionInput,
    TemplateFragment,
    NodeKind,
    acreate_template,
    afind,
    TemplateInput,
)
from fakts.fakts import Fakts
from fluss_next.api.schema import (
    FlowFragment,
)
from reaktion_next.utils import infer_kind_from_graph
from rekuest_next.widgets import StringWidget
from rekuest_next.structures.default import get_default_structure_registry
from rekuest_next.structures.registry import StructureRegistry
from pydantic import BaseModel, Field
from .utils import convert_flow_to_definition
from rekuest_next.agents.extension import AgentExtension

logger = logging.getLogger(__name__)
from rekuest_next.definition.registry import DefinitionRegistry
from rekuest_next.actors.actify import reactify
from rekuest_next.actors.base import ActorTransport
from rekuest_next.collection.collector import Collector


class ReaktionExtension(BaseModel):
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )
    definition_registry: DefinitionRegistry = Field(default_factory=DefinitionRegistry)
    extension_name: str = "reaktion"

    async def should_cleanup_on_init(self):
        return False

    async def aspawn_actor_from_template(
        self,
        template: TemplateFragment,
        passport: Passport,
        transport: ActorTransport,
        agent: "BaseAgent",
        collector: "Collector",
    ) -> Optional[Actor]:
        x = template
        assert "flow" in x.params, "Template is not a flow"

        t = await aget_flow(id=x.params["flow"])

        return FlowActor(
            flow=t,
            is_generator=x.node.kind == NodeKind.GENERATOR,
            passport=passport,
            transport=transport,
            definition=x.node,
            agent=agent,
            collector=collector,
        )

    async def aretrieve_registry(self):
        return self.definition_registry
